@@ -1,3 +1,4 @@
+import functools
 import json
 import logging
 import re
@@ -9,10 +10,22 @@ import fire
 from termcolor import colored
 
 from .cybrex_ai import CybrexAI
+from .exceptions import QdrantStorageNotAvailableError
 
 
-def create_snippet(chunk):
-    return chunk['document_id'] + ': ' + chunk['text']
+def exception_handler(func):
+    @functools.wraps(func)
+    async def wrapper_func(*args, **kwargs):
+        try:
+            return await func(*args, **kwargs)
+        except QdrantStorageNotAvailableError as e:
+            print(
+                f"{colored('INFO', 'red')}: Cannot connect to Qdrant: {e.info}\n"
+                f"{colored('HINT', 'yellow')}: Launch qdrant using `docker run -p 6333:6333 -p 6334:6334 qdrant/qdrant`",
+                file=sys.stderr,
+            )
+            raise
+    return wrapper_func
 
 
 class CybrexCli:
@@ -25,42 +38,28 @@ class CybrexCli:
                 document = json.loads(document)
                 await self.cybrex.upsert_documents([document])
 
-    async def export_chunks(self, query: str, output_path: str, n_documents: int = 10):
-        """
-        Store STC text chunks in ZIP archive
-
-        :param query: query to STC
-        :param output_path: where to store result
-        :param n_documents: the number of chunks to extract
-        """
-        async with self.cybrex as cybrex:
-            print(f"{colored('Q', 'green')}: {query}")
-            await cybrex.export_chunks(
-                query=query,
-                output_path=output_path,
-                n_documents=n_documents
-            )
-
+    @exception_handler
     async def chat_doc(self, document_query: str, query: str, n_chunks: int = 5, minimum_score: float = 0.5):
         """
-        Ask a question about content of source_document identified by DOI.
+        Ask a question about content of document identified by DOI.
 
-        :param document_query: query that returns unique source_document
-        :param query: Text query to the source_document
+        :param document_query: query that returns unique document
+        :param query: Text query to the document
         :param n_chunks: the number of chunks to extract
             more means more tokens to use and more precision in answer
         """
         async with self.cybrex as cybrex:
             print(f"{colored('Document', 'green')}: {document_query}")
             print(f"{colored('Q', 'green')}: {query}")
-            answer, _ = await cybrex.chat_document(
+            cybrex_response = await cybrex.chat_document(
                 document_query,
                 query,
                 n_chunks,
                 minimum_score=minimum_score,
             )
-            print(f"{colored('A', 'green')}: {answer}")
+            print(f"{colored('A', 'green')}: {cybrex_response.answer}")
 
+    @exception_handler
     async def chat_sci(
         self,
         query: str,
@@ -71,41 +70,48 @@ class CybrexCli:
         """
         Ask a general questions
 
-        :param query: text query to the source_document
+        :param query: text query to the document
         :param n_chunks: the number of chunks to extract
             more means more tokens to use and more precision in answer
         :param n_documents: the number of chunks to extract
             more means more tokens to use and more precision in answer
+        :param minimum_score:
         """
         async with self.cybrex as cybrex:
             print(f"{colored('Q', 'green')}: {query}")
-            answer, chunks = await cybrex.chat_science(
+            cybrex_response = await cybrex.chat_science(
                 query=query,
                 n_chunks=n_chunks,
                 n_documents=n_documents,
                 minimum_score=minimum_score,
             )
-            answer = re.sub(r'\(DOI: ([^)]+)\)', r'(https://doi.org/\g<1>)', answer)
-            documents = await self.cybrex.get_documents_from_chunks(chunks)
-            documents = [
-                f'{document.get("doi") or document.get("metadata", {}).get("isbns")}: {document["title"]}'
-                for document in documents
-            ]
-            sources = '\n'.join(documents)
+            answer = re.sub(r'\(DOI:\s*([^)]+)\)', r'(https://doi.org/\g<1>)', cybrex_response.answer)
+            references = []
+            visited = set()
+            for chunk in cybrex_response.chunks:
+                field, value = chunk.document_id.split(':', 1)
+                document_id = f'{field}:{value}'
+                if document_id in visited:
+                    continue
+                visited.add(document_id)
+                references.append(f'{document_id}: {chunk.title}')
+            references = '\n'.join(references)
             print(f"{colored('A', 'green')}: {answer}")
-            print(f"{colored('References', 'green')}:\n{textwrap.indent(sources, ' - ')}")
+            print(f"{colored('References', 'green')}:\n{textwrap.indent(references, ' - ')}")
 
+    @exception_handler
     async def sum_doc(self, document_query: str):
         """
-        Summarization of the source_document
+        Summarization of the document
 
-        :param document_query: query that returns unique source_document
+        :param document_query: query that returns unique document
         """
         async with self.cybrex as cybrex:
             print(f"{colored('Document', 'green')}: {document_query}")
-            answer, _ = await cybrex.summarize_document(document_query)
-            print(f"{colored('Summarization', 'green')}: {answer}")
+            cybrex_response = await cybrex.summarize_document(document_query)
+            print(f"{colored('Summarization', 'green')}: {cybrex_response.answer}")
 
+    @exception_handler
     async def semantic_search(
         self,
         query: str,
@@ -123,18 +129,19 @@ class CybrexCli:
         """
         async with self.cybrex as cybrex:
             print(f"{colored('Q', 'green')}: {query}")
-            chunks = await cybrex.semantic_search(
+            scored_chunks = await cybrex.semantic_search(
                 query=query,
                 n_chunks=n_chunks,
                 n_documents=n_documents,
                 minimum_score=minimum_score
             )
-            snippets = [
-                ' - ' + create_snippet(chunk)
-                for chunk in chunks
-            ]
-            sources = '\n'.join(snippets)
-            print(f"{colored('Sources', 'green')}:\n{sources}")
+            references = []
+            for scored_chunk in scored_chunks:
+                field, value = scored_chunk.chunk.document_id.split(':', 1)
+                document_id = f'{field}:{value}'
+                references.append(f' - {document_id}: {scored_chunk.chunk.title}\n   {scored_chunk.chunk.text}')
+            references = '\n'.join(references)
+            print(f"{colored('References', 'green')}:\n{references}")
 
 
 def cybrex_cli(debug: bool = False):
@@ -148,8 +155,6 @@ def cybrex_cli(debug: bool = False):
         'add-all-chunks': cybrex.add_all_documents,
         'chat-doc': cybrex.chat_doc,
         'chat-sci': cybrex.chat_sci,
-        'export-chunks': cybrex.export_chunks,
-        'import-chunks': cybrex.cybrex.import_chunks,
         'semantic-search': cybrex.semantic_search,
         'sum-doc': cybrex.sum_doc,
         'write-config': cybrex.cybrex.ensure_config,
